/* colorifer.cc
* This file is part of Colorifer
* Copyright (c) 2002,2003,2004 by Stanislav Ievlev
*
* This file is covered by the GNU General Public License,
* which should be included with colorifer as the file COPYING.
*/

#include <cstring>
#include <iostream>
#include <fstream>

#include <signal.h>
#include <locale.h>


#include <boost/scoped_ptr.hpp>
#include <boost/bind.hpp>

#include <subst.hh>
#include <pipeprg.hh>

#include <ncurses/assert.hh>

#ifndef CONFIGDIR
#define CONFIGDIR "../data/"
#endif

using namespace colorifer;

boost::scoped_ptr<pipe_prg>	program; /**< Program we run under pipe, we parse output of this program */

/**
 * after program termination restore terminal colors to default values
 * and send same signal to the child process,
 * no any additional actions, just try to fix system state
 * @param sig number of signal
 */
static void
restore_default_colors_handler(int sig)
{
	if (isatty(STDOUT_FILENO))
	{
		ncurses::otstream out(STDOUT_FILENO);
		out.reset_colors();
		out.set_attr();
	}
	if (isatty(STDERR_FILENO))
	{
		ncurses::otstream out(STDOUT_FILENO);
		out.reset_colors();
		out.set_attr();
	}

	if (program.get()) program->kill(sig);

	_exit (128 + sig);
}

/**
 * handler of SIGSTOP signal
 * @param sig number of signal
 */
static void
sigtstp_handler (int)
{
	signal (SIGTSTP, sigtstp_handler);

	if (isatty(STDOUT_FILENO))
	{
		ncurses::otstream out(STDOUT_FILENO);
		out.reset_colors();
		out.set_attr();
	}
	if (isatty(STDERR_FILENO))
	{
		ncurses::otstream out(STDOUT_FILENO);
		out.reset_colors();
		out.set_attr();
	}

	raise (SIGSTOP);
}
/**
 * calculate exit code of the main program
 * @return main program's exit code
 */
int rc()
{
	if (program.get())
	{
		program->wait(); //wait when child finish it's work
		return program->rc();
	}
	else
		return EXIT_SUCCESS;
}

/** analog of basename */
std::string basic_name(const std::string& name)
{
	std::string::size_type pos = name.rfind('/');
	return (pos == std::string::npos) ? name : name.substr(pos + 1, std::string::npos);
}

/**
 * execute child, process child's output
 * this version color output only if stdout and stderr real terminals
 * @param argv standart *argv[] structure
 * @exception Ing::internal_error
 */
void execute(char *argv[])
{
	subst s;
	bool do_work = isatty(STDOUT_FILENO) && isatty(STDERR_FILENO);

	extern const char *__progname;
	const char *real_name = getenv("COLORIFER_NAME") ? getenv("COLORIFER_NAME") : __progname;
	const std::string config_dir = getenv("COLORIFER_CONFIGDIR") ? getenv("COLORIFER_CONFIGDIR") :CONFIGDIR;
	argv[0] = strdup(real_name); //hide name

	if (do_work)
	{
		const std::string	configname(config_dir + "/"+basic_name(real_name));
		std::ifstream is(configname.c_str());
		if (!is)
			do_work = false; //rollback if cannot open config
		else
			s.load(is); //load pattern definitions from the config
	}

	const char *prog = getenv("COLORIFER_REDIRECT");
	//FIXME: try to get redirect name from config
	if (!prog) throw std::runtime_error("no redirect defined");

	if (do_work)
	{
		try
		{
			program.reset( new pipe_prg(prog, argv) );
			ncurses::otstream c_out(STDOUT_FILENO);
			ncurses::otstream c_err(STDERR_FILENO);

			get_lines(*program,
				  boost::bind(&subst::process, &s, _1, boost::ref(c_out)),
				  boost::bind(&subst::process, &s, _1, boost::ref(c_err)));
		}
		catch (ncurses::runtime_error&) //exec if we cannot initialize curses enviroment
		{
			run(prog, argv);
		}
	}
	else
		run(prog, argv);
}

int main(int, char *argv[])
{
	int retcode = EXIT_SUCCESS;

	//Restore default color before exiting
	signal(SIGINT, restore_default_colors_handler);
	signal(SIGTERM, restore_default_colors_handler);
	signal(SIGQUIT, restore_default_colors_handler);
	signal(SIGPIPE, restore_default_colors_handler);
	signal(SIGTSTP, sigtstp_handler);

	setlocale(LC_ALL, "");

	try
	{
		execute(argv); //run subprocess
		retcode += rc(); //fetch retcode from subprocess
	}
	catch (std::runtime_error& e)
	{
		std::cerr << "fatal: " << e.what() << std::endl;
		retcode = EXIT_FAILURE;
	}
	catch (std::exception& e)
	{
		std::cerr << "exception: " << e.what() << std::endl;
		retcode = EXIT_FAILURE;
	}
	catch (...)
	{
		std::cerr << "unknown exception\n";
		retcode = EXIT_FAILURE;
	}

	return retcode;
}
